using System.Collections;
using System.Collections.Generic;
using System.Linq;
using Leap.Unity;
using MeshVR;
using SimpleJSON;
using UnityEngine;

namespace CheesyFX
{
    public class FreeTheLegs : MVRScript
    {
        private bool isRunning;
        private Atom atom;
        private List<List<JSONClass>> snapshots = new List<List<JSONClass>>(5);
        private List<List<FreeControllerV3>> limbs = new List<List<FreeControllerV3>>(5);
        private JSONStorableFloat undoSteps = new JSONStorableFloat("Undo steps", 5f, 0f, 20f);
        private JSONStorableFloat footJointDriveX = new JSONStorableFloat("FootJointDriveTargetX", -65f, -65f, 40f);
        private JSONStorableFloat toeJointDriveX = new JSONStorableFloat("ToeJointDriveTargetX", 45f, -40f, 75f);
        private JSONStorableFloat freeThighRotationSpring = new JSONStorableFloat("Free thigh rotation spring", 1000f, 300f, 2000f);
        private JSONStorableBool elbowPositionDefault = new JSONStorableBool("Default elbow position state 'On'", true);
        private JSONStorableBool kneePositionDefault = new JSONStorableBool("Default knee position state 'On'", true);
        
        public override void Init()
        {
	        SuperController.singleton.ClearMessages();

            var button = CreateButton("Left Arm");
            button.button.onClick.AddListener(() => FreeLimb(2));
            button = CreateButton("Right Arm", true);
            button.button.onClick.AddListener(() => FreeLimb(3));
            button = CreateButton("Left Leg");
            button.button.onClick.AddListener(() => FreeLimb(0));
            button = CreateButton("Right Leg", true);
            
            button.button.onClick.AddListener(Undo);
            CreateToggle(elbowPositionDefault);
            CreateToggle(kneePositionDefault);
            CreateSlider(freeThighRotationSpring, true);
            CreateSlider(footJointDriveX);
            CreateSlider(toeJointDriveX, true);
            button.button.onClick.AddListener(() => FreeLimb(1));
            button = CreateButton("Undo", true);
            var slider = CreateSlider(undoSteps);
            slider.slider.wholeNumbers = true;
            
            RegisterFloat(freeThighRotationSpring);
            RegisterFloat(footJointDriveX);
            RegisterFloat(toeJointDriveX);
            RegisterBool(elbowPositionDefault);
            RegisterBool(kneePositionDefault);
        }
        
        public void OnBindingsListRequested(List<object> bindings)
        {
            bindings.Add(new Dictionary<string, string>
            {
                { "CheesyFX", "FreeTheFeet" }
            });
            
            bindings.Add(new JSONStorableAction("Free selected limb", () => FreeLimb()));
            bindings.Add(new JSONStorableAction("Default selected limb", ApplyDefaults));
            bindings.Add(new JSONStorableAction("Lock controller in place", LockInPlace));
            bindings.Add(new JSONStorableAction("Undo", Undo));
        }

        private int GetId()
        {
	        int id = -1;
	        var c = SuperController.singleton.GetSelectedController();
	        if (c == null || c == c.containingAtom.mainController) return -1;
	        if (c.name.StartsWith("l")) id = 0;
	        else if (c.name.StartsWith("r")) id = 1;
	        else return -1;
	        if (c.name.Contains("Thigh") || c.name.Contains("Knee") || c.name.Contains("Foot") ||
	            c.name.Contains("Toe")){}
	        else if (c.name.Contains("Arm") || c.name.Contains("Elbow") || c.name.Contains("Hand"))
		        id += 2;
	        else return -1;
	        return id;
        }

        private List<FreeControllerV3> GetLimb(int id)
        {
	        List<FreeControllerV3> limb = new List<FreeControllerV3>();
	        string side = id % 2 == 0 ? "l" : "r";
	        
	        if(id < 2)
	        {
		        limb.Add(atom.freeControllers.FirstOrDefault(x => x.name == side + "ThighControl"));
		        limb.Add(atom.freeControllers.FirstOrDefault(x => x.name == side + "KneeControl"));
		        limb.Add(atom.freeControllers.FirstOrDefault(x => x.name == side + "FootControl"));
		        limb.Add(atom.freeControllers.FirstOrDefault(x => x.name == side + "ToeControl"));
	        }
	        else
	        {
		        limb.Add(atom.freeControllers.FirstOrDefault(x => x.name == side + "ArmControl"));
		        limb.Add(atom.freeControllers.FirstOrDefault(x => x.name == side + "ElbowControl"));
		        limb.Add(atom.freeControllers.FirstOrDefault(x => x.name == side + "HandControl"));
	        }
			
	        snapshots.Insert(0, limb.Select(c => c.GetJSON()).ToList());
	        limbs.Insert(0, limb);
	        if (snapshots.Count > undoSteps.val)
	        {
		        snapshots.RemoveLast();
		        limbs.RemoveLast();
	        }
	        return limb;
        }

        private void Undo()
        {
	        if (isRunning || snapshots.Count == 0) return;
	        for (int i=0; i<limbs[0].Count; i++)
	        {
		        limbs[0][i].RestoreFromJSON(snapshots[0][i]);
	        }
	        // lastIds.RemoveAt(0);
	        limbs.RemoveAt(0);
	        snapshots.RemoveAt(0);
        }

        private void FreeLimb(int id = -1)
        {
	        if (id == -1)
	        {
		        id = GetId();
		        if (id == -1) return;
	        }
            atom = SuperController.singleton.GetSelectedAtom();
            if (atom == null || isRunning) return;
            GetLimb(id);
            StartCoroutine(IEnumFreeLimb());
        }
        
        public IEnumerator IEnumFreeLimb()
        {
	        isRunning = true;
			SuperController.singleton.freezeAnimationToggle.isOn = true;

			List<FreeControllerV3> limb = limbs[0];
			bool leg = limb.Count == 4;
			float timer = 0f;
			
			List<Quaternion> rotations = new List<Quaternion>();
			Quaternion delta0;
			Vector3 angles;
			
			if(leg)
			{
				angles = limb[2].followWhenOff.localRotation.eulerAngles;
				if (angles.x > 180f) angles.x -= 360f;
				if (angles.y > 180f) angles.y -= 360f;
				if (angles.z > 180f) angles.z -= 360f;
				// SuperController.LogMessage(angles.ToString());
				
				limb[0].currentRotationState = FreeControllerV3.RotationState.On;
				limb[1].currentRotationState = FreeControllerV3.RotationState.On;
				limb[1].currentPositionState = FreeControllerV3.PositionState.Off;
	        
				limb[2].currentRotationState = FreeControllerV3.RotationState.Off;
				limb[2].currentPositionState = FreeControllerV3.PositionState.Off;
				limb[3].jointRotationDriveXTarget = toeJointDriveX.val;
				limb[3].currentRotationState = FreeControllerV3.RotationState.Off;
				limb[3].currentPositionState = FreeControllerV3.PositionState.Off;

				limb[2].jointRotationDriveXTarget = -angles.x;
				limb[2].jointRotationDriveYTarget = -angles.z;
				limb[2].jointRotationDriveZTarget = -angles.y;
				
			}
			else
			{
				limb[0].currentRotationState = FreeControllerV3.RotationState.On;
				limb[1].currentRotationState = FreeControllerV3.RotationState.On;
				limb[1].currentPositionState = FreeControllerV3.PositionState.Off;
				limb[2].currentRotationState = FreeControllerV3.RotationState.Off;
				limb[2].currentPositionState = FreeControllerV3.PositionState.Off;

				angles = limb[2].followWhenOff.localRotation.eulerAngles;
				// SuperController.LogMessage(angles.ToString());
				if (angles.x > 180f) angles.x -= 360f;
				if (angles.y > 180f) angles.y -= 360f;
				if (angles.z > 180f) angles.z -= 360f;
				limb[2].jointRotationDriveXTarget = -angles.z;
				limb[2].jointRotationDriveYTarget = -angles.y;
				limb[2].jointRotationDriveZTarget = angles.x;
			}
			for(int i=0; i<limb.Count; i++)
			{
				var c = limb[i];
				rotations.Add(c.followWhenOffRB.transform.rotation);
				c.transform.rotation = c.followWhenOff.rotation;
				c.RBHoldRotationSpring = i == 0 ? 500f : 300f;
			}

			yield return new WaitForSeconds(.5f);
			while(timer < 2f){
				for(int i=0; i<3; i++)
				{
					delta0 = rotations[i] * Quaternion.Inverse(limb[i].followWhenOffRB.transform.rotation);
					limb[i].transform.rotation = delta0 * limb[i].transform.rotation;
				}
				timer += .5f;
				yield return new WaitForSeconds(.5f);
			}
			SuperController.singleton.freezeAnimationToggle.isOn = false;
			isRunning = false;
        }

        private void LockInPlace()
        {
	        var c = SuperController.singleton.GetSelectedController();
	        if (c == null || c == c.containingAtom.mainController) return;
	        limbs.Insert(0, new List<FreeControllerV3>{c});
	        snapshots.Insert(0, new List<JSONClass>{c.GetJSON()});
	        if (snapshots.Count > undoSteps.val)
	        {
		        snapshots.RemoveLast();
		        limbs.RemoveLast();
	        }
	        var transform = c.transform;
	        transform.position = c.followWhenOff.position;
	        transform.rotation = c.followWhenOff.rotation;
	        c.currentPositionState = FreeControllerV3.PositionState.Lock;
	        c.currentRotationState = FreeControllerV3.RotationState.Lock;
	        // c.jointRotationDriveXTarget = 0f;
	        // c.jointRotationDriveYTarget = 0f;
	        // c.jointRotationDriveZTarget = 0f;
        }

        private void ApplyDefaults()
        {
	        int id = GetId();
	        if (id == -1) return;
	        atom = SuperController.singleton.GetSelectedAtom();
	        if (atom == null || isRunning) return;
	        var limb = GetLimb(id);
	        if(id < 2) ApplyLegDefaults(limb);
	        else ApplyArmDefaults(limb);
        }

        private void ApplyLegDefaults(List<FreeControllerV3> limb)
        {
	        limb[2].jointRotationDriveXTarget = footJointDriveX.val;
	        limb[2].GetFloatJSONParam("jointDriveYTarget").SetValToDefault();
	        limb[2].GetFloatJSONParam("jointDriveZTarget").SetValToDefault();
	        
	        for (int i = 0; i < 3; i++)
	        {
		        if(i>0) limb[i].transform.position = limb[i].followWhenOff.position;
		        limb[i].transform.rotation = limb[i].followWhenOff.rotation;
	        }
	        
	        if(kneePositionDefault.val) limb[1].currentPositionState = FreeControllerV3.PositionState.On;
	        else limb[1].currentPositionState = FreeControllerV3.PositionState.Off;
	        limb[1].currentRotationState = FreeControllerV3.RotationState.Off;
	        limb[2].currentPositionState = FreeControllerV3.PositionState.On;
	        limb[2].currentRotationState = FreeControllerV3.RotationState.On;
	        
	        limb[3].currentPositionState = FreeControllerV3.PositionState.Off;
	        limb[3].currentRotationState = FreeControllerV3.RotationState.Off;
	        limb[3].jointRotationDriveXTarget = toeJointDriveX.val;
	        limb[3].GetFloatJSONParam("jointDriveYTarget").SetValToDefault();
	        limb[3].GetFloatJSONParam("jointDriveZTarget").SetValToDefault();
        }

        private void ApplyArmDefaults(List<FreeControllerV3> limb)
        {
	        limb[2].GetFloatJSONParam("jointDriveXTarget").SetValToDefault();
	        limb[2].GetFloatJSONParam("jointDriveYTarget").SetValToDefault();
	        limb[2].GetFloatJSONParam("jointDriveZTarget").SetValToDefault();
	        
	        for (int i = 0; i < 3; i++)
	        {
		        if(i>0) limb[i].transform.position = limb[i].followWhenOff.position;
		        limb[i].transform.rotation = limb[i].followWhenOff.rotation;
	        }
	        
	        if(elbowPositionDefault.val) 
		        limb[1].currentPositionState = FreeControllerV3.PositionState.On;
	        else limb[1].currentPositionState = FreeControllerV3.PositionState.Off;
	        limb[1].currentRotationState = FreeControllerV3.RotationState.Off;
	        limb[2].currentPositionState = FreeControllerV3.PositionState.On;
	        limb[2].currentRotationState = FreeControllerV3.RotationState.On;
        }
    }
}